/**********************************************************************
**  uart_test.c
**	Test UART with random strings
**	Version 1.0
**
**  Copyright (C) 2017 H.Poetzl
**
**	This program is free software: you can redistribute it and/or
**	modify it under the terms of the GNU General Public License
**	as published by the Free Software Foundation, either version
**	2 of the License, or (at your option) any later version.
**
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <time.h>

#include <fcntl.h>
#include <stropts.h>
#include <linux/termios.h>


static unsigned seed = 0;
static unsigned max = 128;

static char rnd[1024] = { 0 };
static char buf[1024] = { 0 };


static inline
void	timestamp(struct timespec *ts)
{
	if (clock_gettime(CLOCK_REALTIME, ts) == -1) {
	    fprintf(stderr,
		"clock_gettime(): %s\n", strerror(errno));
	}
}

int	readn(int fd, char *buf, unsigned len, unsigned max)
{
	struct timespec now;
	unsigned cnt, tot = 0;

	while (tot < len) {
	    cnt = read(fd, buf, max);
	    if (cnt == -1) {
		fprintf(stderr,
		    "read(): %s\n", strerror(errno));
	    } else {
		timestamp(&now);

		fprintf(stdout,
	    	    "[%lu.%09lu]\t%u\t%*.*s\n", 
		    now.tv_sec, now.tv_nsec, 
		    cnt, cnt, cnt, buf);

		tot += cnt;
		buf += cnt;
		max -= cnt;
	    }

	    if (tot < len) {
		fprintf(stderr,
		    "short read %u/%u (%u)\n", tot, len, cnt);
	    }
	}
	return tot;
}

int	main(int argc, char *argv[])
{
	char *dev = "/dev/ttyPS1";
	unsigned long baud = 10000000;
	struct termios2 tio;
	int i, fd, res;

	if (argc > 1) {
	    dev = argv[1];
	    if (argc > 2) {
		baud = strtol(argv[2], NULL, 0);
		if (argc > 3) {
		    seed = strtol(argv[3], NULL, 0);
		}
	    }
	}

	srandom(seed);

	for (i=0; i<sizeof(rnd); i++)
	    rnd[i] = (random() % 42)+48;


	fd = open(dev, O_RDWR);
	if (fd == -1) {
	    fprintf(stderr,
		"error opening >%s<.\n%s\n",
		dev, strerror(errno));
	    exit(1);
	}

	res = ioctl(fd, TCGETS2, &tio);
	if (res == -1) {
	    fprintf(stderr,
		"ioctl(TCGETS2): %s\n", strerror(errno));
	    exit(2);
	}

	tio.c_cflag &= ~PARENB;
	tio.c_cflag &= ~CSTOPB;
	tio.c_cflag &= ~CSIZE;
	tio.c_cflag &= ~CBAUD;;
	tio.c_cflag |= CS8;
	tio.c_cflag |= CLOCAL;
	tio.c_cflag |= BOTHER;

	tio.c_iflag = ICANON;
	tio.c_oflag &= ~OPOST;

	tio.c_ispeed = baud;
	tio.c_ospeed = baud;

	tio.c_cc[VTIME] = 10;
	tio.c_cc[VMIN] = 0;

	res = ioctl(fd, TCSETS2, &tio);
	if (res == -1) {
	    fprintf(stderr,
		"ioctl(TCSETS2): %s\n", strerror(errno));
	    exit(4);
	}

	res = ioctl(fd, TCFLSH, TCIOFLUSH);
	if (res == -1) {
	    fprintf(stderr,
		"ioctl(TCFLSH): %s\n", strerror(errno));
	    exit(4);
	}

	for (i=0; i<10000; i++) {
	    struct timespec now;
	    unsigned pos, cnt;

	    cnt = (random() % (max-1)) + 1;
	    pos = random() % (sizeof(rnd) - cnt);

	    timestamp(&now);

	    res = write(fd, &rnd[pos], cnt);
	    if (res == -1) {
		fprintf(stderr,
		    "write: %s\n", strerror(errno));
	    } else if (res < cnt) {
		fprintf(stderr,
		    "short write: %u/%u\n", res, cnt);

		fprintf(stdout,
	    	    "[%lu.%09lu]\t%u\t%*.*s\n", 
		    now.tv_sec, now.tv_nsec, 
		    res, res, res, &rnd[pos]);
	    } else {
		fprintf(stdout,
	    	    "[%lu.%09lu]\t%u\t%*.*s\n", 
		    now.tv_sec, now.tv_nsec, 
		    cnt, cnt, cnt, &rnd[pos]);
	    }

	    res = readn(fd, buf, cnt, sizeof(buf));
	}

	close(fd);

	exit(0);
}


